#include <sql.h>
#include <sqlext.h>
#include <memory.h>
#include <stdio.h>

#define TABLE_NAME "BUG103287"

typedef struct _hconnection {
    SQLHENV henv;
    SQLHDBC hdbc;
    SQLHSTMT hstmt;
} hconnection_t;

static hconnection_t hconnection;

static void extract_error(char *fn, SQLHANDLE handle, SQLSMALLINT type) {
    SQLINTEGER i = 0;
    SQLINTEGER NativeError;
    SQLCHAR SQLState[ 7 ];
    SQLCHAR MessageText[256];
    SQLSMALLINT TextLength;
    SQLRETURN ret;

    do {
	ret = SQLGetDiagRec(type, handle, ++i, SQLState, &NativeError,
		MessageText, sizeof(MessageText), &TextLength);
	if (SQL_SUCCEEDED(ret)) {
	    printf("%s: %s:%ld:%ld:%s\n", fn,
		    SQLState, (long) i, (long) NativeError, MessageText);
	}
    }
    while( ret == SQL_SUCCESS );
}

#define CHECK_ERROR(_a, _b, _c, _d) \
    if (_a != SQL_SUCCESS) { \
	printf("%s -> %d\n", _b, _a);\
	extract_error(_b, _c, _d); \
	if (!SQL_SUCCEEDED(_a)) { \
	    goto function_exit; \
	}\
    }

#define CHECK_ERROR_WITH_COMMAND(_a, _b, _c, _d) \
    printf("%s -> %d\n", _b, _a);\
if (_a != SQL_SUCCESS) { \
    extract_error(_b, _c, _d); \
    if (!SQL_SUCCEEDED(_a)) { \
	goto function_exit; \
    }\
}


static void connect(char *dsn) {
    SQLCHAR     loutstr[500];
    SQLSMALLINT loutstrlen;
    SQLRETURN   lret; /* ODBC API return status */
    SQLCHAR     lconnect_str[500];
    SQLCHAR     buffer1[100];
    SQLCHAR     buffer2[100];
    /* Allocate an environment handle */
    lret = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &hconnection.henv);
    CHECK_ERROR(lret, "SQLAllocHandle(SQL_HANDLE_ENV)", hconnection.henv, SQL_HANDLE_ENV);
    /* We want ODBC 3 support */
    lret = SQLSetEnvAttr(hconnection.henv, SQL_ATTR_ODBC_VERSION, (void *) SQL_OV_ODBC3, 0);
    CHECK_ERROR(lret, "SQLSetEnvAttr(SQL_OV_ODBC3)", hconnection.henv, SQL_HANDLE_ENV);
    /* Allocate a connection handle */
    lret = SQLAllocHandle(SQL_HANDLE_DBC, hconnection.henv, &hconnection.hdbc);
    CHECK_ERROR(lret, "SQLAllocHandle(SQL_HANDLE_DBC)", hconnection.henv, SQL_HANDLE_ENV);
    /* Connect to the DSN mydsn */
    sprintf((char *)lconnect_str, "DSN=%s;", dsn);
    lret = SQLDriverConnect(hconnection.hdbc, NULL, lconnect_str, SQL_NTS,
	    loutstr, sizeof(loutstr), &loutstrlen,
	    SQL_DRIVER_NOPROMPT);
    CHECK_ERROR(lret, "SQLDriverConnect()", hconnection.hdbc, SQL_HANDLE_DBC);
    lret = SQLSetConnectAttr(hconnection.hdbc, SQL_ATTR_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF, SQL_NTS);
    CHECK_ERROR(lret, "SQLSetConnectAttr(SQL_AUTOCOMMIT_OFF)", hconnection.hdbc, SQL_HANDLE_DBC);
    SQLGetInfo(hconnection.hdbc, SQL_DRIVER_NAME, buffer1, 100, NULL);
    SQLGetInfo(hconnection.hdbc, SQL_DRIVER_VER, buffer2, 100, NULL);
    printf("ODBC Driver %s - version %s\n", buffer1, buffer2);
    SQLGetInfo(hconnection.hdbc, SQL_DBMS_NAME, buffer1, 100, NULL);
    SQLGetInfo(hconnection.hdbc, SQL_DBMS_VER, buffer2, 100, NULL);
    printf("ODBC DBMS %s - version %s\n", buffer1, buffer2);
function_exit:
    lret = lret;
}

static void disconnect(void) {
    SQLRETURN lret = SQLDisconnect(hconnection.hdbc);
    CHECK_ERROR(lret, "SQLDisconnect()", hconnection.hdbc, SQL_HANDLE_DBC);
function_exit:
    lret = lret;
}

static int create_table(void) {
    SQLCHAR command[100];
    sprintf((char *)command, "CREATE TABLE %s(A VARCHAR(10) UNIQUE, B VARCHAR(10))", TABLE_NAME);
    printf("Creating table %s\n", TABLE_NAME);
    SQLRETURN lret = SQLAllocHandle(SQL_HANDLE_STMT, hconnection.hdbc, &hconnection.hstmt);
    CHECK_ERROR(lret, "SQLAllocHandle(SQL_HANDLE_STMT)", hconnection.hdbc, SQL_HANDLE_DBC);
    lret = SQLPrepare(hconnection.hstmt, command, SQL_NTS);
    CHECK_ERROR(lret, "SQLPrepare(SQL_HANDLE_STMT)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLExecute(hconnection.hstmt);
    CHECK_ERROR(lret, "SQLExecute()", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLFreeStmt(hconnection.hstmt, SQL_CLOSE);
    CHECK_ERROR(lret, "SQLFreeStmt()", hconnection.hstmt, SQL_HANDLE_STMT);
function_exit:
    return lret;
}

static int drop_table(void) {
    SQLCHAR command[100];
    sprintf((char *)command, "DROP TABLE %s", TABLE_NAME);
    printf("Dropping table %s\n", TABLE_NAME);
    SQLRETURN lret = SQLAllocHandle(SQL_HANDLE_STMT, hconnection.hdbc, &hconnection.hstmt);
    CHECK_ERROR(lret, "SQLAllocHandle(SQL_HANDLE_STMT)", hconnection.hdbc, SQL_HANDLE_DBC);
    lret = SQLPrepare(hconnection.hstmt, command, SQL_NTS);
    CHECK_ERROR(lret, "SQLPrepare(SQL_HANDLE_STMT)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLExecute(hconnection.hstmt);
    CHECK_ERROR(lret, "SQLExecute()", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLFreeStmt(hconnection.hstmt, SQL_CLOSE);
    CHECK_ERROR(lret, "SQLFreeStmt()", hconnection.hstmt, SQL_HANDLE_STMT);
function_exit:
    return lret;
}

static int insert_records(void) {
    printf("Inserting records into table %s\n", TABLE_NAME);
    SQLCHAR command[100];
    sprintf((char *)command, "INSERT INTO %s VALUES(?,?)", TABLE_NAME);
    SQLRETURN lret = SQLAllocHandle(SQL_HANDLE_STMT, hconnection.hdbc, &hconnection.hstmt);
    CHECK_ERROR(lret, "SQLAllocHandle(SQL_HANDLE_STMT)", hconnection.hdbc, SQL_HANDLE_DBC);
    lret = SQLPrepare(hconnection.hstmt, command, SQL_NTS);
    /* Add 'A','A' */
    CHECK_ERROR(lret, "SQLPrepare(SQL_HANDLE_STMT)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLBindParameter(hconnection.hstmt, 1,
	    SQL_PARAM_INPUT, SQL_C_CHAR,
	    SQL_VARCHAR, 100, 0,
	    "A",
	    0,
	    NULL);
    CHECK_ERROR(lret, "SQLBindParameter(1)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLBindParameter(hconnection.hstmt, 2,
	    SQL_PARAM_INPUT, SQL_C_CHAR,
	    SQL_VARCHAR, 100, 0,
	    "A",
	    0,
	    NULL);
    CHECK_ERROR(lret, "SQLBindParameter(2)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLExecute(hconnection.hstmt);
    CHECK_ERROR(lret, "SQLExecute()", hconnection.hstmt, SQL_HANDLE_STMT);
    /* Add 'B','B' */
    CHECK_ERROR(lret, "SQLPrepare(SQL_HANDLE_STMT)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLBindParameter(hconnection.hstmt, 1,
	    SQL_PARAM_INPUT, SQL_C_CHAR,
	    SQL_VARCHAR, 100, 0,
	    "B",
	    0,
	    NULL);
    CHECK_ERROR(lret, "SQLBindParameter(1)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLBindParameter(hconnection.hstmt, 2,
	    SQL_PARAM_INPUT, SQL_C_CHAR,
	    SQL_VARCHAR, 100, 0,
	    "B",
	    0,
	    NULL);
    CHECK_ERROR(lret, "SQLBindParameter(2)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLExecute(hconnection.hstmt);
    CHECK_ERROR(lret, "SQLExecute()", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLFreeStmt(hconnection.hstmt, SQL_CLOSE);
    CHECK_ERROR(lret, "SQLFreeStmt()", hconnection.hstmt, SQL_HANDLE_STMT);
function_exit:
    return lret;
}

static SQLLEN lnumber_of_returned_rows = 1;
static SQLLEN rows_fetched;
static SQLCHAR buffers[101][10];
static SQLLEN lengths[10];

static int update_record() {
    SQLCHAR command[100];
    sprintf((char *)command, "SELECT A, B FROM %s WHERE A = 'A' FOR UPDATE", TABLE_NAME);
    SQLRETURN lret = SQLAllocHandle(SQL_HANDLE_STMT, hconnection.hdbc, &hconnection.hstmt);
    CHECK_ERROR(lret, "SQLAllocHandle(SQL_HANDLE_STMT)", hconnection.hdbc, SQL_HANDLE_DBC);
    lret = SQLSetStmtAttr(hconnection.hstmt, SQL_ATTR_CONCURRENCY,
	    (SQLPOINTER) SQL_CONCUR_LOCK,
	    0);
    lret = SQLSetStmtAttr(hconnection.hstmt,
	    SQL_ATTR_ROW_ARRAY_SIZE,
	    (SQLPOINTER) lnumber_of_returned_rows,
	    0);
    CHECK_ERROR(lret, "SQLSetStmtAttr(SQL_ATTR_ROW_ARRAY_SIZE)", hconnection.hstmt, SQL_HANDLE_STMT);
    rows_fetched = 0L;
    lret = SQLSetStmtAttr(hconnection.hstmt, SQL_ATTR_ROWS_FETCHED_PTR,
	    &rows_fetched,0);
    CHECK_ERROR(lret, "SQLSetStmtAttr(SQL_ATTR_ROWS_FETCHED_PTR)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLBindCol(hconnection.hstmt, 1, SQL_C_CHAR, buffers[0],
	    100,
	    &lengths[0]);
    CHECK_ERROR(lret, "SQLBindCol(1)", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLBindCol(hconnection.hstmt, 2, SQL_C_CHAR, buffers[1],
	    100,
	    &lengths[1]);
    CHECK_ERROR(lret, "SQLBindCol(2)", hconnection.hstmt, SQL_HANDLE_STMT);

    lret = SQLPrepare(hconnection.hstmt, command, SQL_NTS);
    CHECK_ERROR(lret, "SQLPrepare(SQL_HANDLE_STMT)", hconnection.hstmt, SQL_HANDLE_STMT);

    lret = SQLExecute(hconnection.hstmt);
    CHECK_ERROR(lret, "SQLExecute()", hconnection.hstmt, SQL_HANDLE_STMT);
    lret = SQLFetchScroll(hconnection.hstmt, SQL_FETCH_NEXT, 0);
    if (lret == SQL_NO_DATA) {
	printf("Record not found\n");
    } else {
	CHECK_ERROR(lret, "SQLFetchScroll(SQL_FETCH_NEXT)", hconnection.hstmt,  SQL_HANDLE_STMT);
	// Change the value of the primary key to be that of another row
	buffers[0][0] = 'B';
	// THIS IS THE PROBLEMATIC STATEMEMT
	SQLRETURN lret = SQLSetPos(hconnection.hstmt, 1, SQL_UPDATE, SQL_LOCK_NO_CHANGE);
	CHECK_ERROR_WITH_COMMAND(lret, "SQLSetPos(SQL_UPDATE, SQL_LOCK_NO_CHANGE)", hconnection.hstmt, SQL_HANDLE_STMT);
    }
function_exit:
    return lret;
}

static void commit() {
    SQLRETURN   lret;
    lret =  SQLEndTran(SQL_HANDLE_DBC, hconnection.hdbc, SQL_COMMIT);
    CHECK_ERROR(lret, "SQLEndTran(SQL_COMMIT)", hconnection.hdbc, SQL_HANDLE_DBC);
function_exit:
    lret = lret;
}

int main(int argc, char **argv) {
    if (argc != 3) {
	fprintf(stderr, "Usage: mariabug create/drop/add/update DSN\n");
	return 1;
    }
    char *command = argv[1];
    if (strcmp(command, "create") == 0) {
	connect(argv[2]);
	create_table();
    } else if (strcmp(command, "drop") == 0) {
	connect(argv[2]);
	drop_table();
    } else if (strcmp(command, "insert") == 0) {
	connect(argv[2]);
	insert_records();
    } else if (strcmp(command, "update") == 0) {
	connect(argv[2]);
	update_record();
    } else {
	fprintf(stderr, "Usage: mariabug create/drop/add/update DSN\n");
	return 1;
    }
    commit();
    disconnect();
    return 0;
}
